// libTorrent - BitTorrent library
// Copyright (C) 2005-2011, Jari Sundell
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// In addition, as a special exception, the copyright holders give
// permission to link the code of portions of this program with the
// OpenSSL library under certain conditions as described in each
// individual source file, and distribute linked combinations
// including the two.
//
// You must obey the GNU General Public License in all respects for
// all of the code used other than OpenSSL.  If you modify file(s)
// with this exception, you may extend this exception to your version
// of the file(s), but you are not obligated to do so.  If you do not
// wish to do so, delete this exception statement from your version.
// If you delete this exception statement from all source files in the
// program, then also delete it here.
//
// Contact:  Jari Sundell <jaris@ifi.uio.no>
//
//           Skomakerveien 33
//           3185 Skoppum, NORWAY

#ifndef LIBTORRENT_CHUNK_UTILS_H
#define LIBTORRENT_CHUNK_UTILS_H

#include <vector>

#include <torrent/common.h>
#include <torrent/download.h>

namespace torrent {

class ChunkList;

struct vm_mapping {
  void* ptr;
  uint64_t length;
};

// Change to ChunkList* when that becomes part of the public API.

std::vector<vm_mapping> chunk_list_mapping(Download* download) LIBTORRENT_EXPORT;

struct chunk_info_result {
  Download download;

  uint32_t chunk_index;
  uint32_t chunk_offset;

  const char* file_path;
  uint64_t file_offset;

  // void* chunk_begin;
  // void* chunk_end;
  
  // int prot;
};

chunk_info_result chunk_list_address_info(void* address) LIBTORRENT_EXPORT;

}

#endif
